/*
 * dpkg - main program for package management
 * divert.c - Contains two diversion functions
 *
 * @author Sam Dunne
 * @version 0.1
 * @date 28/06/2011
 *
 * This is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <assert.h>
#include <fcntl.h>
#include <ctype.h>
#include <string.h>
#include <unistd.h>
#include <stdarg.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdbool.h>

#include <dpkg/i18n.h>
#include <dpkg/dpkg.h>
#include <dpkg/dpkg-db.h>
#include <dpkg/parsedump.h>

#include "divert.h"


/**
 * Parse the field and value into the package being constructed.
 * Called once for each field in the stanza.
 * Linked list is built within this function.
 */
static void
divert_parse_field(struct parsedb_state *ps, struct field_state *fs, 
				   void *parse_obj)
{
	const char divert_from[11] 	= "Divert-From";
	const char divert_to[10]	= "Divert-To";

	struct diversion_ctor **head = parse_obj;
	
	/**
	 * Allocate nodes
	 */
	if(!*head) {
		*head = (struct diversion_ctor *)
				malloc(sizeof(struct diversion_ctor));
		(*head)->next = NULL;
	}


	if (strncasecmp(divert_from, fs->fieldstart, fs->fieldlen) == 0 ) {
		/**
		 * FIXME : Check if value exists in struct. If not then return
		 * value to struct otherwise skip.
		 */
		strcpy((*head)->from, fs->valuestart);
	}

	else if (strncasecmp(divert_to, fs->fieldstart, fs->fieldlen) == 0 ) {
		/**
		 * FIXME : Check if value exists in struct. If not then return
		 * value to struct otherwise skip.
		 */
		strcpy((*head)->to, fs->valuestart);
	}

}

/**
 * Called once for each file.
 * Calls parse_stanza() which uses divert_parse_field()
 */
struct diversion_ctor
 *diversion_parse_ci(const char *filename)
{
	/**
	 * For the file.
	 */
	int fd;

	/**
	 * Pointer to our structure
	 */
  	struct diversion_ctor *diversion_result = NULL, 
						  **tail = &diversion_result;
	
	/**
	 * FIXME: Still have to allocate the diversion struct pointer
	 */

	/**
	 * This is the data pointer
	 */
  	char *data;

	/**
	 * File size information.
	 */
  	struct stat st;

	/**
	 * Hold file information.
	 */
  	struct parsedb_state ps;

  	/**
  	 * Tracks what fields have been seen.
  	 */
  	struct field_state fs;

	/** 
	 * Some initialization 
	 */
  	ps.filename = filename;
	
	/** 
	 * Set line number to 0 
	 */
  	ps.lno = 0;	
	
	ps.pkg = NULL;
	
	/** 
	 * Initialize the structure 
	 */
	memset(&fs, 0, sizeof(fs)); 


	/**
	 * Attempt to open file and check if successful.
	 */
  	fd = open(filename, O_RDONLY);

  	/**
  	 * Return error if not opened successfully.
  	 */
  	if (fd == -1) ohshite(_("failed to open package info file `%.255s' for reading"),filename);

	/**
	 * Ensure file is closed again on error.
	 */
  	push_cleanup(cu_closefd, ~ehflag_normaltidy, NULL, 0, 1, &fd);

	/**
	 * Grab file size so we now how big a buffer to allocate.
	 */
  	if (fstat(fd, &st) == -1)
    		ohshite(_("can't stat package info file `%.255s'"),filename);

	/**
	 * Check is size is greater than 0.
	 */
  	if (st.st_size > 0) {

	#ifdef USE_MMAP
		/**
		 * If we're using mmap, we don't have to allocate a buffer at all
		 * Just map the file directly into memory.
		 */
		ps.dataptr = mmap(NULL, st.st_size, PROT_READ, MAP_SHARED, fd, 0);
		if (ps.dataptr == MAP_FAILED)
			ohshite(_("can't mmap package info file `%.255s'"),filename);

	#else
		/**
		 * Otherwise, allocate a buffer the size of the file and read the 
		 * full file contents into the buffer at one go.
		 */
		ps.dataptr = m_malloc(st.st_size);

		if (fd_read(fd, ps.dataptr, st.st_size) < 0)
			ohshite(_("reading package info file '%.255s'"), filename);

	#endif
		/**
		 * Initialize our pointers for use: one pointing at the start of the
		 * buffer which we'll increment as we walk the lines, another 
		 * pointing to the very end of the buffer so we know when we've gone
		 * too far.
		 */
		data = ps.dataptr;
		ps.endptr = ps.dataptr + st.st_size;

  	} else {
  		/**
  		 * If file size is less than or equal to zero set everything to
  		 * NULL.
  		 */
		data = ps.dataptr = ps.endptr = NULL;
  	}

  	/**
  	 * Loop per package.
  	 */
  	for (;;) {

		/**
		 * If parse stanza returns false break
		 * Otherwise this will return the parsed data to the structure
		 * that we need for diversions.
		 */
		if (!parse_stanza(&ps, &fs, divert_parse_field, tail))
			break;
			

		/**
		 * FIXME : Here is the perfect place to fill in the "To:" field
		 * if one doesn't exist in the file.
		 */

		if (parse_EOF(&ps))
			break;
		
		tail = &((*tail)->next);
  	}

	/**
	 * Free our buffer /if/ it was allocated instead of mmapped.
	 */
  	if (data != NULL) {

	#ifdef USE_MMAP
    		munmap(data, st.st_size);

	#else
    		free(data);

	#endif
  	}

  	varbuf_destroy(&fs.value);

  	/**
  	 * Now that we've made it to the end of the function without tripping an
	 * error, we remove our error handler.
	 */
  	pop_cleanup(ehflag_normaltidy);

  	/**
  	 * Close the file.
  	 */
  	if (close(fd)) ohshite(_("failed to close after read: `%.255s'"),filename);
	
	/**
	 * Set last item in the list as NULL
	 */
	tail = NULL;

	/**
	 * Return 'result' to our structure
	 */
	return diversion_result;
}


void
calculate_diversion_list(struct diversion_ctor *old, struct diversion_ctor *new,
                         struct diversion_ctor **added, struct diversion_ctor **removed)
{
	/**
	 * Set temporary pointers
	 */
	struct diversion_ctor *old_temp;
	struct diversion_ctor *new_temp;
	struct diversion_ctor *struct_removed;
	struct diversion_ctor *struct_added;

   	/**
   	 * Initialize the return value
   	 */
   	*removed = NULL;
   	*added   = NULL;

    	/**
    	 * Search the old list with the new list
    	 */
    	for( old_temp = old; old_temp != NULL; old_temp = old_temp->next ) {

        	for( new_temp = new; new_temp!=NULL; new_temp = new_temp->next ) {

            		/**
            		 * Here if both elements match nothing needs to be done so we break
            		 */
            		if( strcmp(old_temp->from, new_temp->from ) == 0) {
                		break;
            		}
        	}

        	/**
        	 * If 'found' has remained at zero it means that the diversion no longer exists
        	 */
        	if( new_temp == NULL ) {
        		struct_removed = (struct diversion_ctor *)
				  		malloc(sizeof(struct diversion_ctor));
        		strcpy(struct_removed->from, old_temp->from);
            		strcpy(struct_removed->to, old_temp->to);
            		struct_removed->next = *removed;
			*removed = struct_removed;
        	}
    	}

    	/**
    	 * Search the old list with the new list
    	 */
    	for( new_temp = new; new_temp != NULL; new_temp = new_temp->next ) {

        	for( old_temp = old; old_temp != NULL; old_temp = old_temp->next ) {

            		/**
            		 * Here if both elements match nothing needs to be done so we break
            		 */
            		if( strcmp(new_temp->from, old_temp->from ) == 0) {
                		break;
            		}
        	}

        	/**
        	 * If 'found' has remained at zero it means that the diversion no longer exists
        	 */
        	if( old_temp == NULL ) {
        		struct_added = (struct diversion_ctor *)
				  	      malloc(sizeof(struct diversion_ctor));
            		strcpy(struct_added->from, new_temp->from);
            		strcpy(struct_added->to, new_temp->to);
            		struct_added->next = *added;
			*added = struct_added;
        	}
    	}
}

/** TEST BLOCK */
int main(void)
{
    	/*
    	 * Declare structure stuff.
		 */
    	 
    	struct diversion_ctor *old_diversions;
    	struct diversion_ctor *new_diversions;
    	struct diversion_ctor *diversions_added;
    	struct diversion_ctor *diversions_removed;
	
		push_error_context();

    	/**
    	 * Allocate.
		 */
    	 
    	old_diversions     = (struct diversion_ctor *)
    			      malloc(sizeof(struct diversion_ctor));
    	new_diversions     = (struct diversion_ctor *)
    			      malloc(sizeof(struct diversion_ctor));
    	diversions_added   = (struct diversion_ctor *)
    			      malloc(sizeof(struct diversion_ctor));
    	diversions_removed = (struct diversion_ctor *)
    			      malloc(sizeof(struct diversion_ctor));

    	/**
    	 * Parse and calculate diversions.
		 */
    	 
    	old_diversions = diversion_parse_ci("div_test/old/diversions");
	new_diversions = diversion_parse_ci("div_test/new/diversions");
	calculate_diversion_list(old_diversions, new_diversions, &diversions_added, &diversions_removed);

    	/*
    	 * PRINTING BLOCK
		 */
    	 
    	printf("\n\nSTART LINKED LIST: OLD\n");
    	while(old_diversions != NULL) {
        	printf(" From: %s | To: %s \n", old_diversions->from, old_diversions->to);
        	old_diversions = old_diversions->next;
    	}
    	printf("END LINKED LIST\n\n");


    	printf("START LINKED LIST: NEW\n");
    	while(new_diversions != NULL) {
        	printf(" From: %s | To: %s \n", new_diversions->from, new_diversions->to);
        	new_diversions = new_diversions->next;
    	}
    	printf("END LINKED LIST\n\n");


    	printf("START LINKED LIST ADDED\n");
    	while(diversions_added != NULL) {
        	printf(" From: %s | To: %s \n", diversions_added->from, diversions_added->to);
        	diversions_added = diversions_added->next;
    	}
    	printf("END LINKED LIST\n\n");



    	printf("START LINKED LIST: REMOVED\n");
    	while(diversions_removed != NULL) {
        	printf(" From: %s | To: %s \n", diversions_removed->from, diversions_removed->to);
        	diversions_removed = diversions_removed->next;
    	}
    	printf("END LINKED LIST\n\n");

    return 0;
}

